/****************************************************************************
 ****************************************************************************/

#include "jwm.h"

static ClientNode *nodes = NULL;
static ClientNode *activeClient = NULL;

static void Reparent(ClientNode *np);
 
#ifdef USE_MWM_HINTS
static PropMwmHints *GetMwmHints(Window w);
#endif
static void ReadClientProtocols(ClientNode *np);
static void SendClientMessage(ClientNode *np, Atom type, Atom message);
static void GetBorderOffsets(ClientNode *np, int *north, int *west);
static void PlaceWindow(ClientNode *np, int alreadyMapped);
static void Gravitate(ClientNode *np, int negate);

/****************************************************************************
 ****************************************************************************/
void AddClientWindow(Window w, int alreadyMapped) {
	XWindowAttributes attr;
#ifdef USE_MWM_HINTS
	PropMwmHints *mhints;
#endif
	ClientNode *np;

	if(w == trayWindow || w == rootWindow) {
		return;
	}

	np = Allocate(sizeof(ClientNode));
	np->next = nodes;
	nodes = np;
	np->window = w;

	XFetchName(display, w, &np->name);

	ReadClientHints(np);

	XGetWindowAttributes(display, w, &attr);
	np->x = attr.x;
	np->y = attr.y;
	np->width = attr.width;
	np->height = attr.height;
	np->colorMap = attr.colormap;
	np->isActive = 0;

	np->borderFlags = BORDER_OUTLINE | BORDER_TITLE
		| BORDER_MIN | BORDER_MAX | BORDER_CLOSE;
	np->borderAction = BA_NONE;
#ifdef USE_MWM_HINTS
	mhints = GetMwmHints(w);
	if(mhints) {
		if(mhints->flags & MWM_HINTS_DECORATIONS
			&& !(mhints->decorations & MWM_DECOR_ALL)) {
			if(!(mhints->decorations & MWM_DECOR_BORDER)) {
				np->borderFlags &= ~BORDER_OUTLINE;
			}
			if(!(mhints->decorations & MWM_DECOR_TITLE)) {
				np->borderFlags &= ~BORDER_TITLE;
			}
		}
	}
#endif

	ReadClientProtocols(np);

	Reparent(np);

	PlaceWindow(np, alreadyMapped);

	np->ignoreUnmap = 1 + alreadyMapped;
	np->isMapped = 1;
	SetDefaultCursor(np->window);
	XMapWindow(display, np->window);
	XMapRaised(display, np->parent);

	DrawBorder(np);

	AddClientToTray(np);

	SendConfigureEvent(np);

}

/****************************************************************************
 ****************************************************************************/
void PlaceWindow(ClientNode *np, int alreadyMapped) {
	int north, west;

	GetBorderOffsets(np, &north, &west);

	if(!(np->sizeFlags & PPosition) && !alreadyMapped) {
		np->x = west + rootWidth / 2 - np->width / 2;
		np->y = north + rootHeight / 2 - np->height / 2;
	} else {
		Gravitate(np, 0);
	}

	XMoveWindow(display, np->parent, np->x - west, np->y - north);
}

/****************************************************************************
 ****************************************************************************/
void GetBorderOffsets(ClientNode *np, int *north, int *west) {

	*north = 0;
	*west = 0;

	if(np->borderFlags & BORDER_OUTLINE) {
		*west = BORDER_WEST;
		if(np->borderFlags & BORDER_TITLE) {
			*north = BORDER_NORTH;
		} else {
			*north = BORDER_SOUTH;
		}
	}

}

/****************************************************************************
 ****************************************************************************/
void Gravitate(ClientNode *np, int negate) {
	int north, south, west;
	int northDelta, westDelta;

	GetBorderOffsets(np, &north, &west);
	if(north) {
		south = BORDER_SOUTH;
	} else {
		south = 0;
	}

	northDelta = 0;
	westDelta = 0;
	switch(np->gravity) {
	case 1: /* NorthWest */
		northDelta = -south;
		westDelta = -west;
		break;
	case 2: /* North */
		northDelta = -south;
		break;
	case 3: /* NorthEast */
		northDelta = -south;
		westDelta = west;
		break;
	case 4: /* West */
		westDelta = -west;
		break;
	case 5: /* Center */
		northDelta = (north + south) / 2;
		westDelta = west;
		break;
	case 6: /* East */
		westDelta = west;
		break;
	case 7: /* SouthWest */
		northDelta = north;
		westDelta = -west;
		break;
	case 8: /* South */
		northDelta = north;
		break;
	case 9: /* SouthEast */
		northDelta = north;
		westDelta = west;
		break;
	default: /* Static */
		break;
	}

	if(negate) {
		np->x -= westDelta;
		np->y -= northDelta;
	} else {
		np->x += westDelta;
		np->y += northDelta;
	}

}

/****************************************************************************
 ****************************************************************************/
void MinimizeClient(ClientNode *np) {
	np->isMapped = 0;
	np->ignoreUnmap += 2;
	XUnmapWindow(display, np->window);
	XUnmapWindow(display, np->parent);
	DrawTray();
}

/****************************************************************************
 ****************************************************************************/
void RestoreClient(ClientNode *np) {
	XMapWindow(display, np->window);
	XMapRaised(display, np->parent);
	np->isMapped = 1;
}

/****************************************************************************
 ****************************************************************************/
void MaximizeClient(ClientNode *np) {

	np->x = BORDER_WEST;
	np->y = BORDER_NORTH;
	np->width = rootWidth - BORDER_EAST - BORDER_WEST;
	np->height = rootHeight - BORDER_NORTH - BORDER_SOUTH;

	XMoveResizeWindow(display, np->parent,
		np->x - BORDER_EAST, np->y - BORDER_NORTH,
		np->width + BORDER_EAST + BORDER_WEST,
		np->height + BORDER_NORTH + BORDER_SOUTH);
	XMoveResizeWindow(display, np->window, BORDER_EAST,
		BORDER_NORTH, np->width, np->height);
	SendConfigureEvent(np);
}

/****************************************************************************
 ****************************************************************************/
void FocusClient(ClientNode *np) {

	if(!np->isMapped) {
		return;
	}

	if(activeClient && activeClient != np) {
		activeClient->isActive = 0;
		DrawBorder(activeClient);
	}
	np->isActive = 1;
	activeClient = np;
	if(np->colorMap) {
		XInstallColormap(display, np->colorMap);
	}

	if(np->protocols & PROT_TAKE_FOCUS) {
		SendClientMessage(np, wmProtocols, wmTakeFocus);
	}

	DrawBorder(np);
	DrawTray();

	XSetInputFocus(display, np->window, RevertToPointerRoot, CurrentTime);
}

/****************************************************************************
 ****************************************************************************/
void RefocusClient() {
	if(activeClient) {
		FocusClient(activeClient);
	}
}

/****************************************************************************
 ****************************************************************************/
void DeleteClient(ClientNode *np) {
	if(np->protocols & PROT_DELETE) {
		SendClientMessage(np, wmProtocols, wmDelete);
	} else {
		XKillClient(display, np->window);
	}
}

/****************************************************************************
 ****************************************************************************/
void SendClientMessage(ClientNode *np, Atom type, Atom message) {
	XEvent event;
	int status;

	memset(&event, 0, sizeof(event));
	event.xclient.type = ClientMessage;
	event.xclient.window = np->window;
	event.xclient.message_type = type;
	event.xclient.format = 32;
	event.xclient.data.l[0] = message;
	event.xclient.data.l[1] = CurrentTime;

	status = XSendEvent(display, np->window, False, 0, &event);
	if(status == False) {
		fprintf(stderr, "SendClientMessage failed\n");
	}
}

/****************************************************************************
 ****************************************************************************/
void SetShape(ClientNode *np) {
	XRectangle *rectangle;
	int count, order;

	rectangle = XShapeGetRectangles(display, np->window, ShapeBounding,
		&count, &order);
	if(rectangle) {
		if(count > 1) {
			if(np->borderFlags & BORDER_OUTLINE) {
				if(np->borderFlags & BORDER_TITLE) {
					XShapeCombineShape(display, np->parent, ShapeBounding,
						BORDER_EAST, BORDER_NORTH, np->window, ShapeBounding,
						ShapeSet);
				} else {
					XShapeCombineShape(display, np->parent, ShapeBounding,
						BORDER_EAST, BORDER_SOUTH, np->window, ShapeBounding,
						ShapeSet);
				}
			} else {
				XShapeCombineShape(display, np->parent, ShapeBounding,
					0, 0, np->window, ShapeBounding, ShapeSet);
			}
		}
		XFree(rectangle);
	}
}

/****************************************************************************
 ****************************************************************************/
void MoveClient(ClientNode *np) {
	XEvent event;
	int startx, starty;
	int oldx, oldy;
	int doMove = 0;

	if(!(np->borderFlags & BORDER_OUTLINE)) {
		return;
	}

	if(!GrabMouseForMove()) {
		return;
	}

	GetMousePosition(&startx, &starty);
	oldx = np->x;
	oldy = np->y;

	for(;;) {
		XMaskEvent(display, ButtonReleaseMask | PointerMotionMask
			| ExposureMask, &event);

		if(!(GetMouseMask() & (Button1Mask | Button3Mask))) {
			goto StopMove;
		}

		switch(event.type) {
		case Expose:
			HandleExpose(&event.xexpose);
			break;
		case MotionNotify:

			while(XCheckTypedEvent(display, MotionNotify, &event));

			np->x = oldx + event.xmotion.x - startx;
			np->y = oldy + event.xmotion.y - starty;

			if(!doMove && (abs(np->x - oldx) > MOVE_DELTA
				|| abs(np->y - oldy) > MOVE_DELTA)) {
				CreateMoveWindow();
				doMove = 1;
			}

			if(doMove) {
				if(np->borderFlags & BORDER_TITLE) {
					XMoveWindow(display, np->parent, np->x - BORDER_EAST,
						np->y - BORDER_NORTH);
				} else {
					XMoveWindow(display, np->parent, np->x - BORDER_EAST,
						np->y - BORDER_SOUTH);
				}
				UpdateMoveWindow(np->x, np->y);
			}

			break;
		default:
			break;
		}
	}

StopMove:
	XUngrabPointer(display, CurrentTime);

	if(!doMove) {
		np->x = oldx;
		np->y = oldy;
	} else {
		DestroyMoveWindow();
	}

	if(np->borderFlags & BORDER_TITLE) {
		XMoveWindow(display, np->parent, np->x - BORDER_EAST,
			np->y - BORDER_NORTH);
	} else {
		XMoveWindow(display, np->parent, np->x - BORDER_EAST,
			np->y - BORDER_SOUTH);
	}
	SendConfigureEvent(np);

}

/****************************************************************************
 ****************************************************************************/
void DrawOutline(ClientNode *np) {
	int north;

	if(np->borderFlags & BORDER_TITLE) {
		north = BORDER_NORTH;
	} else {
		north = BORDER_SOUTH;
	}

	XDrawRectangle(display, rootWindow, invertedGC,
		np->x - BORDER_EAST, np->y - north,
		np->width + BORDER_EAST + BORDER_WEST,
		np->height + north + BORDER_SOUTH);
}

/****************************************************************************
 ****************************************************************************/
void ResizeClient(ClientNode *np, BorderActionType action) {

	XEvent event;
	int startx, starty;
	int oldx, oldy;
	int oldw, oldh;
	int gwidth, gheight;
	int delta;
	int north;

	if(!(np->borderFlags & BORDER_OUTLINE)) {
		return;
	}

	if(!GrabMouseForResize(action)) {
		return;
	}

	GetMousePosition(&startx, &starty);

	DrawOutline(np);

	oldx = np->x;
	oldy = np->y;
	oldw = np->width;
	oldh = np->height;

	gwidth = (np->width - np->baseWidth) / np->xinc;
	gheight = (np->height - np->baseHeight) / np->yinc;

	CreateResizeWindow();
	UpdateResizeWindow(gwidth, gheight);

	for(;;) {

		XMaskEvent(display, ButtonReleaseMask | PointerMotionMask, &event);

		if(!(GetMouseMask() & Button1Mask)) {
			goto StopResize;
		}

		switch(event.type) {
		case MotionNotify:
			DrawOutline(np);

			if(action & BA_RESIZE_N) {
				delta = (event.xmotion.y - starty) / np->yinc;
				delta *= np->yinc;
				if(oldh - delta >= np->minHeight
					&& oldh - delta <= np->maxHeight) {
					np->height = oldh - delta;
					np->y = oldy + delta;
				}
			}
			if(action & BA_RESIZE_S) {
				delta = (event.xmotion.y - starty) / np->yinc;
				delta *= np->yinc;
				np->height = oldh + delta;
				np->height = Max(np->height, np->minHeight);
				np->height = Min(np->height, np->maxHeight);
			}
			if(action & BA_RESIZE_E) {
				delta = (event.xmotion.x - startx) / np->xinc;
				delta *= np->xinc;
				np->width = oldw + delta;
				np->width = Max(np->width, np->minWidth);
				np->width = Min(np->width, np->maxWidth);
			}
			if(action & BA_RESIZE_W) {
				delta = (event.xmotion.x - startx) / np->xinc;
				delta *= np->xinc;
				if(oldw - delta >= np->minWidth
					&& oldw - delta <= np->maxWidth) {
					np->width = oldw - delta;
					np->x = oldx + delta;
				}
			}

			gwidth = (np->width - np->baseWidth) / np->xinc;
			gheight = (np->height - np->baseHeight) / np->yinc;

			DrawOutline(np);
			UpdateResizeWindow(gwidth, gheight);

			break;
		default:
			break;
		}
	}
StopResize:
	DrawOutline(np);
	XUngrabPointer(display, CurrentTime);

	DestroyResizeWindow();

	if(np->borderFlags & BORDER_TITLE) {
		north = BORDER_NORTH;
	} else {
		north = BORDER_SOUTH;
	}

	XMoveResizeWindow(display, np->parent,
		np->x - BORDER_EAST, np->y - north,
		np->width + BORDER_EAST + BORDER_WEST,
		np->height + north + BORDER_SOUTH);
	XMoveResizeWindow(display, np->window, BORDER_EAST,
		north, np->width, np->height);

	SendConfigureEvent(np);

}

/****************************************************************************
 ****************************************************************************/
void MapClient(XMapEvent *e) {
	ClientNode *np;

	np = FindClientByWindow(e->window);
	if(!np) {
		AddClientWindow(e->window, 0);
	}
}

/****************************************************************************
 ****************************************************************************/
void UnmapClient(XUnmapEvent *e) {
	ClientNode *np;

	np = FindClientByWindow(e->window);
	if(np) {
		if(np->ignoreUnmap) {
			--np->ignoreUnmap;
		} else {
			/* FIXME */
			XUngrabButton(display, AnyButton, AnyModifier, np->window);
			XReparentWindow(display, np->window, rootWindow, np->x, np->y);
			RemoveClient(np);
		}
	}

}

/****************************************************************************
 ****************************************************************************/
void RemoveClient(ClientNode *np) {
	ClientNode *tp;

	if(activeClient == np) {
		activeClient = NULL;
	}

	np->isMapped = 0;
	RemoveClientFromTray(np);

	if(shouldExit) {
		Gravitate(np, 1);
		XUngrabButton(display, AnyButton, AnyModifier, np->window);
		XReparentWindow(display, np->window, rootWindow, np->x, np->y);
	}
	if(np->parent) {
		XFreeGC(display, np->parentGC);
		XDestroyWindow(display, np->parent);
	}

	if(np->name) {
		XFree(np->name);
	}
	if(np->colorMap) {
		XUninstallColormap(display, np->colorMap);
	}

	if(nodes == np) {
		nodes = np->next;
	} else {
		for(tp = nodes; tp; tp = tp->next) {
			if(tp->next == np) {
				tp->next = np->next;
				break;
			}
		}
	}

	Release(np);

}

/****************************************************************************
 ****************************************************************************/
ClientNode *FindClientByWindow(Window w) {
	ClientNode *np;

	for(np = nodes; np; np = np->next) {
		if(np->window == w || np->parent == w) {
			return np;
		}
	}
	return NULL;
}

/****************************************************************************
 ****************************************************************************/
ClientNode *FindClientByParent(Window p) {
	ClientNode *np;

	for(np = nodes; np; np = np->next) {
		if(np->parent == p) {
			return np;
		}
	}
	return NULL;
}

/****************************************************************************
 ****************************************************************************/
void Reparent(ClientNode *np) {
	XSetWindowAttributes attr;
	int x, y, width, height;

	XAddToSaveSet(display, np->window);
	XSelectInput(display, np->window, PropertyChangeMask
		| Button1Mask | EnterWindowMask | ColormapChangeMask);

	XGrabButton(display, AnyButton, AnyModifier, np->window,
		True, ButtonPressMask, GrabModeSync, GrabModeAsync, None, None);

	attr.override_redirect = True;
	attr.event_mask = ButtonPressMask | ExposureMask | PointerMotionMask
		| SubstructureRedirectMask | SubstructureNotifyMask
		| EnterWindowMask;
	attr.background_pixel = borderColor.pixel;

	if(np->borderFlags & BORDER_OUTLINE) {
		x = np->x - BORDER_EAST;
		width = np->width + BORDER_EAST + BORDER_WEST;
		if(np->borderFlags & BORDER_TITLE) {
			y = np->y - BORDER_NORTH;
			height = np->height + BORDER_NORTH + BORDER_SOUTH;
		} else {
			y = np->y - BORDER_SOUTH;
			height = np->height + 2 * BORDER_SOUTH;
		}
	} else {
		x = np->x;
		y = np->y;
		width = np->width;
		height = np->height;
	}

	np->parent = XCreateWindow(display, rootWindow,
		x, y, width, height, 0,
		DefaultDepth(display, 0), InputOutput,
		DefaultVisual(display, 0),
		CWOverrideRedirect | CWBackPixel | CWEventMask, &attr);
	np->parentGC = XCreateGC(display, np->parent, 0, NULL);

#ifdef USE_SHAPE
	if(haveShape) {
		XShapeSelectInput(display, np->window, ShapeNotifyMask);
		SetShape(np);
	}
#endif

	XSetWindowBorderWidth(display, np->window, 0);
	if(np->borderFlags & BORDER_OUTLINE) {
		if(np->borderFlags & BORDER_TITLE) {
			XReparentWindow(display, np->window, np->parent, BORDER_EAST,
				BORDER_NORTH);
		} else {
			XReparentWindow(display, np->window, np->parent, BORDER_EAST,
				BORDER_SOUTH);
		}
	} else {
		XReparentWindow(display, np->window, np->parent, 0, 0);
	}

	SendConfigureEvent(np);
}

/****************************************************************************
 ****************************************************************************/
void SendConfigureEvent(ClientNode *np) {
	XConfigureEvent event;

	event.type = ConfigureNotify;
	event.event = np->window;
	event.window = np->window;
	event.x = np->x;
	event.y = np->y;
	event.width = np->width;
	event.height = np->height;
	event.border_width = 0;
	event.above = None;
	event.override_redirect = False;

	XSendEvent(display, np->window, False, StructureNotifyMask,
		(XEvent*)&event);
}

/****************************************************************************
 ****************************************************************************/
void ReleaseClients() {
	while(nodes) {
		RemoveClient(nodes);
	}
}

/****************************************************************************
 ****************************************************************************/
#ifdef USE_MWM_HINTS
PropMwmHints *GetMwmHints(Window w) {
	Atom type;
	unsigned long itemCount, bytesLeft;
	unsigned char *data;
	int format;

	if(XGetWindowProperty(display, w, mwmHints, 0L, 20L, False,
		mwmHints, &type, &format, &itemCount, &bytesLeft, &data)
		== Success) {
		if(itemCount >= PROP_MOTIF_WM_HINTS_ELEMENTS) {
			return (PropMwmHints*)data;
		}
	}
	return NULL;
}
#endif

/****************************************************************************
 ****************************************************************************/
void ReadClientProtocols(ClientNode *np) {
	long count, x;
	int status;
	unsigned long extra;
	Atom realType;
	int realFormat;
	unsigned char *temp;
	Atom *protocols;

	np->protocols = PROT_NONE;

	status = XGetWindowProperty(display, np->window, wmProtocols, 0, 32,
		False, XA_ATOM, &realType, &realFormat, &count, &extra, &temp);
	protocols = (Atom*)temp;

	if(status != Success || !protocols) {
		return;
	}
	if(count <= 0) {
		XFree(protocols);
		return;
	}

	for(x = 0; x < count; x++) {
		if(protocols[x] == wmDelete) {
			np->protocols |= PROT_DELETE;
		} else if(protocols[x] == wmTakeFocus) {
			np->protocols |= PROT_TAKE_FOCUS;
		}
	}

	XFree(protocols);

}

/****************************************************************************
 ****************************************************************************/
void ReadClientHints(ClientNode *np) {
	XSizeHints hints;
	long temp;

	if(!XGetWMNormalHints(display, np->window, &hints, &temp)) {
		np->sizeFlags = 0;
	} else {
		np->sizeFlags = hints.flags;
	}

	if(np->sizeFlags & PResizeInc) {
		np->xinc = Max(1, hints.width_inc);
		np->yinc = Max(1, hints.height_inc);
	} else {
		np->xinc = 1;
		np->yinc = 1;
	}

	if(np->sizeFlags & PMinSize) {
		np->minWidth = Max(0, hints.min_width);
		np->minHeight = Max(0, hints.min_height);
	} else {
		np->minWidth = 1;
		np->minHeight = 1;
	}

	if(np->sizeFlags & PMaxSize) {
		np->maxWidth =  hints.max_width;
		np->maxHeight = hints.max_height;
		if(np->maxWidth <= 0) {
			np->maxWidth = rootWidth;
		}
		if(np->maxHeight <= 0) {
			np->maxHeight = rootHeight;
		}
	} else {
		np->maxWidth = rootWidth;
		np->maxHeight = rootHeight;
	}

	if(np->sizeFlags & PBaseSize) {
		np->baseWidth = hints.base_width;
		np->baseHeight = hints.base_height;
	} else if(np->sizeFlags & PMinSize) {
		np->baseWidth = np->minWidth;
		np->baseHeight = np->minHeight;
	} else {
		np->baseWidth = 0;
		np->baseHeight = 0;
	}

	if(np->sizeFlags & PWinGravity) {
		np->gravity = hints.win_gravity;
	} else {
		np->gravity = 1;
	}

}


