/****************************************************************************
 ****************************************************************************/

#include "jwm.h"

#ifdef USE_SHAPE
static int shapeEvent;
#endif

static char *CONFIG_FILE = "/.jwmrc";

static void Initialize();
static void ScanWindows();
static void EventLoop();
static void Shutdown();
static void HandleExit();
static void DoExit(int code);

Atom wmState;
Atom wmProtocols;
Atom wmDelete;
Atom wmTakeFocus;
static Atom wmChangeState;
static Atom wmColormaps;
static Atom wmRunning;

static char *configPath = NULL;
static char *displayString = NULL;

/****************************************************************************
 ****************************************************************************/
int main(int argc, char *argv[]) {
	char *temp;
	int x;

	StartDebug();

	temp = getenv("HOME");
	if(temp) {
		configPath = Allocate(strlen(temp) + strlen(CONFIG_FILE) + 1);
		strcpy(configPath, temp);
		strcat(configPath, CONFIG_FILE);
	} else {
		configPath = Allocate(strlen(CONFIG_FILE) + 1);
		strcpy(configPath, CONFIG_FILE);
	}

	for(x = 1; x < argc; x++) {
		if(!strcmp(argv[x], "-v")) {
			DisplayAbout();
			DoExit(0);
		} else if(!strcmp(argv[x], "-h")) {
			DisplayHelp();
			DoExit(0);
		} else if(!strcmp(argv[x], "-p")) {
			ParseConfig(configPath);
			DestroyRootMenu();
			ReleaseFonts();
			DoExit(0);
		} else if(!strcmp(argv[x], "-display") && x + 1 < argc) {
			displayString = argv[++x];
		} else {
			DisplayUsage();
			DoExit(1);
		}
	}

	do {
		shouldExit = 0;
		shouldRestart = 0;
		ParseConfig(configPath);
		Initialize();
		EventLoop();
		Shutdown();
	} while(shouldRestart);
	Release(configPath);
	StopDebug();

	return 0;
}

/****************************************************************************
 ****************************************************************************/
void DoExit(int code) {
	if(configPath) {
		Release(configPath);
	}
	StopDebug();
	exit(code);
}

/****************************************************************************
 ****************************************************************************/
void EventLoop() {
	struct timeval timeout;
	int fd;
	fd_set fds;
	XEvent event;

	fd = XConnectionNumber(display);

	while(!shouldExit) {

		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		timeout.tv_sec = MAIN_DELTA;
		timeout.tv_usec = 0;
		if(select(fd + 1, &fds, NULL, NULL, &timeout) <= 0) {
			UpdateTime();
		}

		while(XCheckMaskEvent(display,
			SubstructureRedirectMask | SubstructureNotifyMask | ExposureMask
			| PropertyChangeMask | EnterWindowMask | ColormapChangeMask
			| ButtonPressMask | PointerMotionMask,
			&event) == True) {

			switch(event.type) {
			case ButtonPress:
				HandleButtonPress(&event.xbutton);
				break;
			case ConfigureRequest:
				HandleConfigureRequest(&event.xconfigurerequest);
				break;
			case MapRequest:
				MapClient(&event.xmap);
				break;
			case EnterNotify:
				HandleEnterNotify(&event.xcrossing);
				break;
			case PropertyNotify:
				HandlePropertyNotify(&event.xproperty);
				break;
			case UnmapNotify:
				UnmapClient(&event.xunmap);
				break;
			case Expose:
				HandleExpose(&event.xexpose);
				break;
			case MotionNotify:
				HandleMotionNotify(&event.xmotion);
				break;
			case ColormapNotify:
				HandleColormapChange(&event.xcolormap);
				break;
			case DestroyNotify:
				HandleDestroyNotify(&event.xdestroywindow);
				break;
			default:
#ifdef USE_SHAPE
				if(haveShape && event.type == shapeEvent) {
					HandleShapeEvent((XShapeEvent*)&event);
				}
#endif
				break;
			}
		}
	}
}

/****************************************************************************
 ****************************************************************************/
void Initialize() {
	XGCValues gcvalues;
	XSetWindowAttributes attr;
	int temp;

	initializing = 1;
	display = XOpenDisplay(displayString);
	if(!display) {
		if(displayString) {
			printf("error: could not open display %s\n", displayString);
		} else {
			printf("error: could not open display\n");
		}
		DoExit(1);
	}

	XSetErrorHandler(ErrorHandler);

	wmState = XInternAtom(display, "WM_STATE", False);
	wmChangeState = XInternAtom(display, "WM_CHANGE_STATE", False);
	wmProtocols = XInternAtom(display, "WM_PROTOCOLS", False);
	wmDelete = XInternAtom(display, "WM_DELETE_WINDOW", False);
	wmTakeFocus = XInternAtom(display, "WM_TAKE_FOCUS", False);
	wmColormaps = XInternAtom(display, "WM_COLORMAP_WINDOWS", False);
	wmRunning = XInternAtom(display, "_JWM_RUNNING", False);
#ifdef USE_MWM_HINTS
	mwmHints = XInternAtom(display, _XA_MWM_HINTS, False);
#endif

	InitializeColors();
	InitializeFonts();
	InitializeCursors();

	rootWindow = RootWindow(display, 0);
	rootWidth = DisplayWidth(display, DefaultScreen(display));
	rootHeight = DisplayHeight(display, DefaultScreen(display));

	attr.event_mask = SubstructureRedirectMask | SubstructureNotifyMask
		| PropertyChangeMask | ColormapChangeMask | ButtonPressMask
		| ButtonReleaseMask;
	XChangeWindowAttributes(display, rootWindow, CWEventMask, &attr);

	LoadImages();

	SetDefaultCursor(rootWindow);

	gcvalues.function = GXinvert;
	gcvalues.subwindow_mode = IncludeInferiors;
	gcvalues.line_width = 2;
	invertedGC = XCreateGC(display, rootWindow, GCFunction
		| GCSubwindowMode | GCLineWidth, &gcvalues);

	signal(SIGTERM, HandleExit);
	signal(SIGINT, HandleExit);
	signal(SIGHUP, HandleExit);

#ifdef USE_SHAPE
	haveShape = XShapeQueryExtension(display, &shapeEvent, &temp);
	if(!haveShape) {
		Debug("No shape extension.");
	}
#endif

	InitializeLoadDisplay();
	CreateTray();

	ScanWindows();

	initializing = 0;

}

/****************************************************************************
 *Scan already existing windows.
 ****************************************************************************/
void ScanWindows() {
	XWindowAttributes attr;
	Window rootReturn, parentReturn, *childrenReturn;
	unsigned int childrenCount;
	unsigned int x;

	XQueryTree(display, rootWindow, &rootReturn, &parentReturn,
		&childrenReturn, &childrenCount);

	for(x = 0; x < childrenCount; x++) {
		XGetWindowAttributes(display, childrenReturn[x], &attr);
		if(attr.override_redirect == False && attr.map_state == IsViewable) {
			AddClientWindow(childrenReturn[x], 1);
		}
	}

	XFree(childrenReturn);
}

/****************************************************************************
 ****************************************************************************/
void HandleExit() {
	signal(SIGTERM, HandleExit);
	signal(SIGINT, HandleExit);
	signal(SIGHUP, HandleExit);
	shouldExit = 1;
}

/****************************************************************************
 ****************************************************************************/
void Shutdown() {
	DestroyRootMenu();
	DestroyLoadDisplay();
	DestroyTray();
	ReleaseClients();
	UnloadImages();
	ReleaseCursors();
	ReleaseFonts();
	ReleaseColors();
	XCloseDisplay(display);
}


