/***************************************************************************
 ***************************************************************************/

#include "jwm.h"

static int ShowMenu(MenuType *menu, MenuType *parent, int x, int y);
static void CreateMenu(MenuType *menu, int x, int y);
static void DestroyMenu(MenuType *menu);
static void DrawMenu(MenuType *menu);

static int MenuLoop(MenuType *menu);
static int UpdateMotion(MenuType *menu, XMotionEvent *event);
static void UpdateMenu(MenuType *menu);
static MenuItemType *GetMenuItem(MenuType *menu, int index);

static void RunCommand(const char *command);
static void DestroyRootMenuHelper(MenuType *menu);

static MenuType *rootMenu = NULL;
static int menuItemHeight;

/***************************************************************************
 ***************************************************************************/
void SetRootMenu(MenuType *m) {
	if(rootMenu) {
		DestroyRootMenu();
	}
	rootMenu = m;
}

/***************************************************************************
 ***************************************************************************/
void DestroyRootMenu() {
	DestroyRootMenuHelper(rootMenu);
	rootMenu = NULL;
}

/***************************************************************************
 ***************************************************************************/
void DestroyRootMenuHelper(MenuType *menu) {
	MenuItemType *np;
	if(menu) {
		while(menu->items) {
			np = menu->items->next;
			if(menu->items->name) {
				Release(menu->items->name);
			}
			if(menu->items->command) {
				Release(menu->items->command);
			}
			if(menu->items->submenu) {
				DestroyRootMenuHelper(menu->items->submenu);
			}
			Release(menu->items);
			menu->items = np;
		}
		Release(menu);
	}
}

/***************************************************************************
 ***************************************************************************/
void ShowRootMenu(int x, int y) {

	if(!rootMenu) {
		return;
	}

	menuItemHeight = menuFont->ascent + menuFont->descent + 1;

	if(!GrabMouseForMenu()) {
		return;
	}

	ShowMenu(rootMenu, NULL, x, y);

	RefocusClient();

}

/***************************************************************************
 ***************************************************************************/
int ShowMenu(MenuType *menu, MenuType *parent, int x, int y) {
	int status;

	CreateMenu(menu, x, y);
	menu->parent = parent;

	status = MenuLoop(menu);

	DestroyMenu(menu);

	if(!parent) {
		XUngrabPointer(display, CurrentTime);
		RefocusClient();
	}

	return status;
}

/***************************************************************************
 * Returns 0 if no selection was made or 1 if a selection was made.
 ***************************************************************************/
int MenuLoop(MenuType *menu) {
	XEvent event;
	MenuItemType *ip;
	int count;

	for(;;) {
		XMaskEvent(display, ButtonPressMask | PointerMotionMask
			| ExposureMask, &event);

		switch(event.type) {
		case Expose:
			HandleExpose(&event.xexpose);
			break;
		case MotionNotify:
			switch(UpdateMotion(menu, &event.xmotion)) {
			case 0: /* no selection */
				break;
			case 1: /* mouse left the menu */
				return 0;
			case 2: /* selection made */
				return 1;
			}
			break;
		case ButtonPress:
			if(menu->currentIndex >= 0) {
				count = 0;
				for(ip = menu->items; ip; ip = ip->next) {
					if(count == menu->currentIndex && ip->command) {
						RunCommand(ip->command);
					}
					++count;
				}
			}
			return 1;
			break;
		default:
			break;
		}
	}
}

/***************************************************************************
 ***************************************************************************/
void CreateMenu(MenuType *menu, int x, int y) {
	MenuItemType *np;
	int temp;

	menu->lastIndex = -1;
	menu->currentIndex = -1;
	menu->width = 4;
	menu->height = 0;
	for(np = menu->items; np; np = np->next) {
		menu->height += menuItemHeight;
		if(np->name) {
			temp = XTextWidth(menuFont, np->name, strlen(np->name));
			if(temp > menu->width) {
				menu->width = temp;
			}
		}
	}
	menu->width += 8;

	if(x + menu->width > rootWidth) {
		x = rootWidth - menu->width;
	}
	temp = y;
	if(y + menu->height > rootHeight - TRAY_HEIGHT - 2) {
		y = rootHeight - TRAY_HEIGHT - menu->height - 2;
	}

	menu->x = x;
	menu->y = y;
	menu->parentOffset = temp - y;

	menu->window = XCreateSimpleWindow(display, rootWindow, x, y,
		menu->width, menu->height, 1, borderOutlineColor.pixel,
		borderColor.pixel);
	XMapRaised(display, menu->window);

	XSync(display, False);

	XSetInputFocus(display, menu->window, RevertToPointerRoot, CurrentTime);
	menu->gc = XCreateGC(display, menu->window, 0, NULL);
	XSetFont(display, menu->gc, menuFont->fid);

	DrawMenu(menu);

}

/***************************************************************************
 ***************************************************************************/
void DestroyMenu(MenuType *menu) {
	XFreeGC(display, menu->gc);
	XDestroyWindow(display, menu->window);
}

/***************************************************************************
 ***************************************************************************/
void DrawMenu(MenuType *menu) {
	MenuItemType *np;
	int y;

	XSetForeground(display, menu->gc, borderColor.pixel);
	XFillRectangle(display, menu->window, menu->gc,
		0, 0, menu->width, menu->height);

	XSetForeground(display, menu->gc, borderTextColor.pixel);
	XSetBackground(display, menu->gc, borderColor.pixel);
	y = menuFont->ascent;
	for(np = menu->items; np; np = np->next) {
		if(np->name) {
			XDrawImageString(display, menu->window, menu->gc, 4, y,
				np->name, strlen(np->name));
		}
		y += menuItemHeight;
	}

}

/***************************************************************************
 * Return values:
 *  0 no selection.
 *  1 mouse left the menu.
 *  2 selection made on a submenu.
 ***************************************************************************/
int UpdateMotion(MenuType *menu, XMotionEvent *event) {
	MenuItemType *ip;
	MenuType *tp;
	int x, y;

	x = event->x - menu->x;
	y = event->y - menu->y;

	/* Update the selection on the current menu */
	if(x > 0 && y > 0 && x < menu->width && y < menu->height) {
		menu->currentIndex = y / menuItemHeight;
	} else {
		for(tp = menu->parent; tp; tp = tp->parent) {
			if(event->subwindow == tp->window) {
				if(y < menu->parentOffset
					|| y > menuItemHeight + menu->parentOffset) {
					return 1;
				}
			}
		}
		menu->currentIndex = -1;
	}
	if(menu->lastIndex != menu->currentIndex) {
		UpdateMenu(menu);
		menu->lastIndex = menu->currentIndex;
	}

	/* If the selected item is a submenu, show it. */
	ip = GetMenuItem(menu, menu->currentIndex);
	if(ip && ip->submenu) {
		if(ShowMenu(ip->submenu, menu, menu->x + menu->width,
			menu->y + menu->currentIndex * menuItemHeight)) {

			/* Item selected; destroy the menu tree. */
			return 2;

		} else {

			/* No selection made; redraw. */
			DrawMenu(menu);
			UpdateMenu(menu);

		}
	}

	return 0;

}

/***************************************************************************
 ***************************************************************************/
void UpdateMenu(MenuType *menu) {
	MenuItemType *ip;
	int yoffset;

	/* Clear the old selection. */
	ip = GetMenuItem(menu, menu->lastIndex);
	if(ip) {
		XSetForeground(display, menu->gc, borderColor.pixel);
		XFillRectangle(display, menu->window, menu->gc,
			0, menu->lastIndex * menuItemHeight, menu->width,
			menuItemHeight + 2);
		XSetForeground(display, menu->gc, borderTextColor.pixel);
		XSetBackground(display, menu->gc, borderColor.pixel);

		if(ip->name) {
			yoffset = menuFont->ascent + menu->lastIndex * menuItemHeight;
			XDrawImageString(display, menu->window, menu->gc, 4, yoffset,
				ip->name, strlen(ip->name));
		}
	}

	/* Highlight the new selection. */
	ip = GetMenuItem(menu, menu->currentIndex);
	if(ip) {
		XSetForeground(display, menu->gc, activeBorderColor.pixel);
		XFillRectangle(display, menu->window, menu->gc,
			0, menu->currentIndex * menuItemHeight, menu->width,
			menuItemHeight);
		XSetForeground(display, menu->gc, activeBorderTextColor.pixel);
		XSetBackground(display, menu->gc, activeBorderColor.pixel);

		if(ip->name) {
			yoffset = menuFont->ascent + menu->currentIndex * menuItemHeight;
			XDrawImageString(display, menu->window, menu->gc, 4, yoffset,
				ip->name, strlen(ip->name));
		}
	}

}

/***************************************************************************
 ***************************************************************************/
MenuItemType *GetMenuItem(MenuType *menu, int index) {
	MenuItemType *ip;

	if(index >= 0) {
		for(ip = menu->items; ip; ip = ip->next) {
			if(!index) {
				return ip;
			}
			--index;
		}
	} else {
		ip = NULL;
	}

	return ip;
}

/***************************************************************************
 ***************************************************************************/
void RunCommand(const char *command) {
	char *displayString;
	char *str;

	if(!command) {
		return;
	}

	if(!strcmp(command, "#exit")) {
		shouldExit = 1;
		return;
	} else if(!strcmp(command, "#restart")) {
		shouldRestart = 1;
		shouldExit = 1;
		return;
	}

	displayString = DisplayString(display);

	if(!fork()) {
		if(!fork()) {
			close(ConnectionNumber(display));

			if(displayString && displayString[0]) {
				str = malloc(strlen(displayString) + 9);
				sprintf(str, "DISPLAY=%s", displayString);
				putenv(str);
			}

			execl(SHELL_NAME, SHELL_NAME, "-c", command, NULL);
			fprintf(stderr, "exec failed\n");
			exit(1);

		}
		exit(0);
	}

	wait(NULL);

}

